/*
 * Symbolic disassembler support for VDA68k
 * Copyright 2012, Adrien Destugues
 *
 * This file is distributed under the terms of the MIT licence
 */
#include "symbols.h"

#include <stdio.h>
#include <stdlib.h>

struct symbol* syms;
struct symbol* comments;
struct symbol* areas;

char* find_symbol(unsigned long addr, unsigned long* offset)
{
	struct symbol* sym;
   
	for(sym = syms; sym != NULL && sym->addr > addr; sym = sym->next)
	{
	}

	if (sym == NULL) {
		*offset = addr;
		return 0;
	}

	*offset = addr - sym->addr;
	return sym->name;
}

void AddSymbol(int addr, const char* name)
{
	struct symbol* newSym;
	struct symbol* prevSym = syms;
	struct symbol* nextSym = NULL;

	newSym = malloc(sizeof(int) + sizeof(struct symbol*) + strlen(name) + 1);
	newSym->addr = addr;
	strcpy(newSym->name, name);

	for(;;)
	{
		if (prevSym == NULL || prevSym->addr < newSym->addr)
		{
			newSym->next = prevSym;
			if (nextSym == NULL)
			{
				syms = newSym;
			} else {
				nextSym->next = newSym;
			}
			break;
		} else {
			nextSym = prevSym;
			prevSym = prevSym->next;
		}
	}
}


void AddArea(int addr, char kind)
{
	struct symbol* newSym;
	struct symbol* prevSym = areas;
	struct symbol* nextSym = NULL;

	newSym = malloc(sizeof(int) + sizeof(struct symbol*) + 1);
	newSym->addr = addr;
	newSym->name[0] = kind;

	for(;;)
	{
		if (prevSym == NULL || prevSym->addr > newSym->addr)
		{
			newSym->next = prevSym;
			if (nextSym == NULL)
			{
				areas = newSym;
			} else {
				nextSym->next = newSym;
			}
			break;
		} else {
			nextSym = prevSym;
			prevSym = prevSym->next;
		}
	}
}


void ReadSyms(const char* file)
{
	FILE* f;
	char name[40];
	int	addr;
	int line = 1;

	syms = NULL;
	comments = NULL;
	areas = NULL;

	f = fopen(file, "r");
	if (f == NULL) return;

	while(!feof(f))
	{

		char kind = fgetc(f);

		switch(kind)
		{
			case 's':
				fscanf(f, "%39s %i\n", name, &addr);
				AddSymbol(addr, name);
				break;
			case 'c':
			case 'b':
			case 'w':
			case 'd':
			case 'a':
				fscanf(f, "%i\n", &addr);
				AddArea(addr, kind);
				break;
			case '#':
			case ';':
				/* Comments in the control file are ignored */
				break;
			default:
				printf("%s:%d: unknown directive\n", file, line);
		}

		++line;
	}

	fclose(f);
}

void freeSyms(void)
{
	while(syms) {
		struct symbol* sym;
		sym	= syms->next;
		free(syms);
		syms = sym;	
	}
}
